# -*- coding: utf-8 -*-
"""Basic methods for selection operations.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2020, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.0.1"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

from numpy import random

def selection_elites_tournament(individuals : list, n : int, tournament_size : int = 5, island=None) -> list:
    elites = []
    for i in range(n):
        selection = selection_elites_random(individuals, tournament_size)
        elites.extend(selection_elites_top_n(selection, 1))
    return elites

def selection_elites_random(individuals : list, n : int, island=None) -> list:
    return random.choice(individuals, size=n).tolist()

def selection_elites_top_n(individuals : list, n : int, desc : bool = True, island=None) -> list:
    """
    A Classic top N selection function, sorted on fitness.

    Args:
        individuals (list): A list of Individuals.
        n (int): Number to select.
        desc (bool): In descending order (default = True).
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Top N Individuals.
    """
    def sortFitness(val):
        return val.fitness

    individuals.sort(key=sortFitness, reverse=desc)

    return individuals[0:n]


# -*- coding: utf-8 -*-
"""Basic methods for crossover operations.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2020, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.0.1"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

import random

def crossover_two_uniform(individuals : list, prob : float = 0.5, island=None) -> list:
    """
    A Classic crossover function, taking a list of 2 individuals and swapping positional genes based on the `prob` strength of crossover.
    Returns these two individuals with modified genomes.

    Args:
        individuals (list): A list (length of 2) of Individuals to perform crossover.
        prob (float): The probability of swapping genes.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Two new Individuals.
    """
    assert len(individuals) > 1, "Not enough individuals given!"
    mother = individuals[0]
    father = individuals[1]
    size = min(len(mother.chromosome), len(father.chromosome))
    for i in range(size):
        if random.random() < prob:
            mother.chromosome[i], father.chromosome[i] = father.chromosome[i], mother.chromosome[i]

    return [mother, father]

def crossover_two_n_point(individuals : list, n_points : int = 1, island=None) -> list:
    """
    Classic crossover method to randomly select N points for crossover.

    Args:
        individuals (list): A list (length of 2) of Individuals to perform crossover.
        prob (float): The probability of swapping genes.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Two new Individuals.
    """
    assert len(individuals) > 1, "Not enough individuals given!"
    mother = individuals[0]
    father = individuals[1]
    size = min(len(mother.chromosome), len(father.chromosome))

    point_cut_list = random.sample(range(1,size-1), n_points)
    point_cut_list.sort()
    point_cut_list.insert(0,0)
    point_cut_list.append(size)
    for i in range(len(point_cut_list)-1):
        b = point_cut_list[i]
        e = point_cut_list[i+1]
        mother.chromosome[b:e], father.chromosome[b:e] = father.chromosome[b:e], mother.chromosome[b:e]

    return [mother, father]


# -*- coding: utf-8 -*-
"""Basic methods for selection operations.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2020, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.0.1"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

def selection_parents_two(individuals : list, island=None):
    """
    Simple function to select two parents at a time, sequentially.

    Args:
        individuals (list): A list of Individuals, specifically the selected "elites".
        island (Island): The Island calling the method (optional, default = None).

    Yields:
        list: Containing the two individuals selected for crossover.
    """
    for parent_1, parent_2 in zip(individuals[::2], individuals[1::2]):
        yield [parent_1, parent_2]

def selection_parents_two_shuffled(individuals : list, island=None):
    raise NotImplementedError("Future implementation")

def selection_parents_n_gram(individuals : list, n : int = 2,island=None):
    raise NotImplementedError("Future implementation")


import unittest
from natural_selection.genetic_algorithms import Gene, Chromosome, Individual
from natural_selection.genetic_algorithms.utils.random_functions import random_int

class TestIndividual(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda chromosome, island, x, y: chromosome[0].value * x + y
        g_1 = Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="test_other", value=4, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)

    def test_birthday(self):
        self.assertEquals(self.ind.age, 0)
        self.ind.birthday()
        self.assertEquals(self.ind.age, 1)
        self.ind.birthday(3)
        self.assertEquals(self.ind.age, 4)

    def test_reset_fitness(self):
        self.ind.fitness = 5
        self.ind.reset_fitness()
        self.assertIsNone(self.ind.fitness)
        self.ind.reset_fitness(fitness=4)
        self.assertEquals(self.ind.fitness, 4)

        self.ind.genetic_code = "true"
        self.ind.reset_fitness(reset_genetic_code=False)
        self.assertEquals(self.ind.genetic_code, "true")
        self.ind.reset_fitness(reset_genetic_code=True)
        self.assertIsNone(self.ind.genetic_code)

    def test_add_gene(self):
        self.ind.add_gene(Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int))
        self.assertEquals(len(self.ind.chromosome), 3)

    def test_evaluate(self):
        f = self.ind.evaluate({'x' : 2, 'y' : 5})
        self.assertEquals(f, 11)
        self.assertNotEquals(f, 10)

    def test_unique_genetic_code(self):
        code = self.ind.unique_genetic_code()
        self.assertIsNotNone(code)


from numpy import random

def random_int(gene):
    return random.randint(low=gene.gene_min, high=gene.gene_max)

def random_int_step(gene):
    stepped_value = random.randint(low=gene.step_lower_bound, high=gene.step_upper_bound)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = random.randint(low=gene.step_lower_bound, high=gene.step_upper_bound)

    return stepped_value

def random_gaussian(gene):
    stepped_value = random.normal(loc=gene.mu, scale=gene.sig)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = random.normal(loc=gene.mu, scale=gene.sig)

    return stepped_value

def random_uniform(gene):
    return random.uniform(low=gene.gene_min, high=gene.gene_max)

def random_gaussian_step(gene):
    stepped_value = gene.value + random.normal(loc=gene.mu, scale=gene.sig)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = gene.value + random.normal(loc=gene.mu, scale=gene.sig)

    return stepped_value

def random_uniform_step(gene):
    stepped_value = gene.value + random.uniform(low=gene.step_lower_bound, high=gene.step_upper_bound)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = gene.value + random.uniform(low=gene.step_lower_bound, high=gene.step_upper_bound)

    return stepped_value

def random_choice(gene):

    if not 'choices' in gene.__dict__:
        raise KeyError("'choices' not defined in this gene, please include a list values!")
    return random.choice(gene.choices)


# -*- coding: utf-8 -*-
"""Basic methods for mutation operations.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2020, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.0.1"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

import random

def mutation_randomize(individual, prob : float = 0.2, island=None):
    """
    A Classic mutation function, changes a gene of the given individual based on the `prob` strength of mutation.

    Args:
        individual: Individual object containing a Genome.
        prob: The probability of randomizing genes.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        Individual: The newly mutated individual.
    """
    for i in range(len(individual.chromosome)):
        if random.random() < prob:
            individual.chromosome[i].randomise()

    return individual


def initialise_population_random(adam, n : int, island=None):
    population = list()

    for i in range(n - 1):
        chromosome = island.create_chromosome([x.randomise_new() for x in adam.chromosome])
        eve = island.create_individual(adam.fitness_function, chromosome=chromosome)
        population.append(eve)

    population.append(adam)

    return population


import copy

def clone_classic(individuals : list, island=None):
    """
    Classic cloning function, making a deep copy of population list.

    Args:
        individuals (list): Population members to copy.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Carbon copy of population.
    """
    return copy.deepcopy(individuals)

class GeneticAlgorithmError(Exception):

    def __init__(self, message : str, exit : bool = False):
        self.message = message
        if exit:
            print(f"GeneticAlgorithmError: {self.message}")
            quit(1)


    def __str__(self):
        return f"GeneticAlgorithmError: {self.message}"